"""This module holds the Account class for the lazuli package.

Copyright 2020 TEAM SPIRIT. All rights reserved.
Use of this source code is governed by a AGPL-style license that can be found in the LICENSE file.
Refer to database.py or the project wiki on GitHub for usage examples.
"""
import mysql.connector as con


class Account:
	"""Account object; models AzureMS accounts.

	Using instance method Lazuli::get_account_by_username(username) or the Lazuli::get_char_by_name(name).account getter
	will create an Account object instance with attributes identical to the account with username "username"
	(or IGN "name" for the latter) in the connected AzureMS-based database.
	This class contains the appropriate getter and setter methods for said attributes.

	Attributes:
		account_id: Integer, representing Primary Key for account - Do NOT set manually
		username: String, representing account username; varchar(64)
		logged_in: Integer, represents login status of the account; int(1) and not bool for some reason
		banned: Integer, represents ban status of the account; int(1) and not bool for some reason
		ban_reason: String, representing account ban reason; text
		nx: Integer, representing the amount of NX Prepaid the user has
		maple_points: Integer, representing the amount of Maple Points the user has
		vp:Integer, representing the amount of Vote Points the user has
		dp: Integer, representing the amount of Donation Points the user has
		char_slots: Integer, representing the number of character slots the user has
	"""
	def __init__(self, account_info, database_config):
		"""Emulates how account object is handled server-sided

		Args:
			account_info: dictionary of user attributes, formatted in AzureMS style
			database_config: dictionary of protected attributes from a Lazuli object
		"""

		self._account_info = account_info
		self._database_config = database_config

		self._account_id = 0  # Primary Key - Do NOT set
		self._username = ""  # varchar(64)
		self._logged_in = 0  # int(1) and not bool for some reason
		self._banned = 0  # int(1) and not bool for some reason
		self._ban_reason = ""  # text
		# self._gm = 0  # There are conflicting values in character and accounts - ignored for now
		self._nx = 0
		self._maple_points = 0
		self._vp = 0
		self._dp = 0
		self._char_slots = 0

		self.init_account_stats()

	def init_account_stats(self):
		"""Given a dictionary of stats from AzureMS's DB we add them to Account object's attributes

		Runs near the end of Account::__init__(account_info, database_config).
		It assigns the account attributes in account_info to their respective protected attributes belonging to
		the Account object instance.
		"""
		self._account_id = self._account_info['id']
		self._username = self._account_info['name']
		self._logged_in = self._account_info['loggedin']
		self._banned = self._account_info['banned']
		self._ban_reason = self._account_info['banreason']
		# self._gm = self._account_info['gm']  # There are conflicting values in character and accounts - ignored for now
		self._nx = self._account_info['nxCash']
		self._maple_points = self._account_info['mPoints']
		self._vp = self._account_info['vpoints']
		self._dp = self._account_info['realcash']  # Best guess - Might be wrong!
		self._char_slots = self._account_info['chrslot']

	@property
	def account_info(self):
		return self._account_info

	@property
	def database_config(self):
		return self._database_config

	@property
	def account_id(self):
		return self._account_id  # Primary Key; DO NOT set

	@property
	def username(self):
		return self._username

	@username.setter
	def username(self, new_name):
		self.set_stat_by_column("name", new_name)
		self._username = new_name

	@property
	def logged_in(self):
		return self._logged_in

	@logged_in.setter
	def logged_in(self, value):
		self.set_stat_by_column("loggedin", value)  # Use with caution!
		self._logged_in = value

	@property
	def banned(self):
		return self._banned

	@banned.setter
	def banned(self, value):
		self.set_stat_by_column("banned", value)  # Use with caution!
		self._banned = value

	@property
	def ban_reason(self):
		return self._ban_reason

	@ban_reason.setter
	def ban_reason(self, value):
		self.set_stat_by_column("banreason", value)
		self._ban_reason = value

	@property
	def nx(self):
		return self._nx

	@nx.setter
	def nx(self, value):
		self.set_stat_by_column("nxCash", value)
		self._nx = value

	def add_nx(self, amount):
		"""Adds the specified amount to the current NX pool

		Args:
			amount: Int, representing the amount of NX to be added to the NX pool
		"""
		new_nx = int(amount) + self.nx
		self.nx = new_nx

	@property
	def maple_points(self):
		return self._nx

	@maple_points.setter
	def maple_points(self, value):
		self.set_stat_by_column("mPoints", value)
		self._maple_points = value

	def add_maple_points(self, amount):
		"""Adds the specified amount to the current Maple Points pool

		Args:
			amount: Int, representing the number of Maple Points to be added to the current pool
		"""
		new_maple_points = int(amount) + self.maple_points
		self.maple_points = new_maple_points

	@property
	def vp(self):
		return self._vp

	@vp.setter
	def vp(self, value):
		self.set_stat_by_column("vpoints", value)
		self._vp = value

	def add_vp(self, amount):
		"""Adds the specified amount to the current VP count

		Args:
			amount: Int, representing the number of vote points to be added to the current count
		"""
		new_vp = int(amount) + self.vp
		self.vp = new_vp

	@property
	def dp(self):
		return self._dp

	@dp.setter
	def dp(self, value):
		self.set_stat_by_column("realcash", value)
		self._dp = value

	def add_dp(self, amount):
		"""Adds the specified amount to the current DP count

		Args:
			amount: Int, representing the number of DPs to be added to the current count
		"""
		new_dp = int(amount) + self.dp
		self.dp = new_dp

	@property
	def char_slots(self):
		return self._char_slots

	@char_slots.setter
	def char_slots(self, value):
		self.set_stat_by_column("chrslot", value)
		self._char_slots = value

	def add_char_slots(self, amount):
		"""Adds the specified amount to the current character slot count

		Args:
			amount: Int, representing the number of slots to be added to the current count
		"""
		new_count = int(amount) + self.char_slots
		self.char_slots = new_count

	def is_online(self):
		"""Checks if the 'loggedin' column is greater than 0 (they are online if > 0)

		Returns:
			Boolean, representing the online status of the account
		"""
		if int(self.logged_in) > 0:
			return True
		return False

	def unstuck(self):
		"""Sets loggedin column in database to 0

		This unstucks the account as server checks loggedin value to decided whether they are "logged in"
		"""
		self.logged_in = 0

	def change_password(self, new_pass):
		"""Changes the current password to the given one.

		WARNING: INHERENTLY UNSAFE
		Azure316 does not hash passwords by default, therefore this function is technically functional.
		As "safe" as any website registration or auto register.

		Args:
			new_pass: string, representing the new password
		"""
		self.set_stat_by_column("password", new_pass)

	def get_stat_by_column(self, column):
		"""Fetches account attribute by column name

		Returns:
			Int or String, representing user attribute queried

		Raises:
			Generic error on failure
		"""
		try:
			return self.account_info[str(column)]
		except Exception as e:
			print("[ERROR] Error trying to obtain the given column for table users.", e)

	def set_stat_by_column(self, column, value):
		"""Sets a account's attributes by column name in database

		Grabs the database attributes provided through the class constructor.
		Uses these attributes to attempt a database connection.
		Attempts to update the field represented by the provided column in the accounts table, with the provided value.
		Not recommended to use this alone, as it won't update the account object which this was used from

		Args:
			value: int or string, representing the value to be set in the database
			column: string, representing the column in the database that is to be updated

		Returns:
			A boolean representing whether the operation was successful.

		Raises:
			SQL Error 2003: Can't cannect to DB
			WinError 10060: No response from DB
			List index out of range: Wrong column name
		"""
		host = self._database_config["host"]
		user = self._database_config["user"]
		password = self._database_config["password"]
		schema = self._database_config["schema"]
		port = self._database_config["port"]

		try:
			database = con.connect(host=host, user=user, password=password, database=schema, port=port)

			cursor = database.cursor(dictionary=True)
			cursor.execute(f"UPDATE accounts SET {column} = '{value}' WHERE id = '{self.account_id}'")
			database.commit()
			print(f"Successfully updated {column} value for user id: {self.account_id}.")
			self._account_info[column] = value  # Update the stats in the dictionary
			database.disconnect()
			return True
		except Exception as e:
			print("[ERROR] Error trying to set stats in database.", e)
			return False



"""This module holds the Character class for the lazuli package.

Copyright 2020 TEAM SPIRIT. All rights reserved.
Use of this source code is governed by a AGPL-style license that can be found in the LICENSE file.
Refer to database.py or the project wiki on GitHub for usage examples.
"""
import mysql.connector as con

from lazuli import JOBS
from lazuli.inventory import Inventory
from lazuli.account import Account
import lazuli.utility as utils


class Character:
    """Character object; models AzureMS characters.

    Using instance method Lazuli::get_char_by_name(name) will create a Character object instance with
    attributes identical to the character with IGN "name" in the connected AzureMS-based database.
    This class contains the appropriate getter and setter methods for said attributes.

    Attributes:
        character_id: Integer, representing Primary Key for Character; int(11)
        account_id: Integer, representing Primary Key for Account (FK); int(11)
        name: String, representing Character IGN; varchar(13)
        level: Integer, representing Character level
        exp: Integer, representing Character EXP; bigint(20)
        strength: Integer, representing Character STR stat pool
        dex: Integer, representing Character DEX stat pool
        luk: Integer, representing Character LUK stat pool
        inte: Integer, representing Character INT stat pool
        max_hp: Integer, representing Character Max HP stat pool
        max_mp: Integer, representing Character Max MP stat pool
        meso: Integer, representing character wealth (aka Meso count)
        job: Integer, representing Job ID of the character
        skin: Integer, representing Skin ID of the character
        gender: Integer, representing Gender ID of the character
        fame: Integer, representing Character fame count
        hair: Integer, representing Hair ID of the character
        face: Integer, representing Face ID of the character
        ap: Integer, representing Character free Ability Points (AP) pool
        map: Integer, representing Map ID of the map that the character is currently in
        bl_slots: Integer, representing Character Buddy List slots
        rebirths: Integer, representing Character rebirth count
        ambition: Integer, representing Character Ambition pool
        insight: Integer, representing Character Insight pool
        willpower: Integer, representing Character Willpower pool
        diligence: Integer, representing Character Diligence pool
        empathy: Integer, representing Character Empathy pool
        charm: Integer, representing Character Charm pool
        honour: Integer, representing Character Honour pool
        mute: String, representing whether a character is chat-banned
    """

    def __init__(self, char_stats, database_config):
        """Emulates how character object is handled server-sided

        Not all character attributes are inherited, as the database table design in AzureMS is quite verbose

        Args:
            char_stats: dictionary of character stats, formatted in AzureMS style
            database_config: dictionary of protected attributes from a Lazuli object
        """
        self._stats = char_stats
        self._database_config = database_config

        self._character_id = 0
        self._account_id = 0
        self._name = ""  # varchar 13
        self._level = 0
        self._exp = 0
        self._strength = 0
        self._dex = 0
        self._luk = 0
        self._inte = 0
        self._max_hp = 0
        self._max_mp = 0
        self._meso = 0
        self._job = 0
        self._skin = 0
        self._gender = 0
        self._fame = 0
        self._hair = 0
        self._face = 0
        self._ap = 0
        self._map = 0
        self._bl_slots = 0
        self._rebirths = 0
        self._ambition = 0
        self._insight = 0
        self._willpower = 0
        self._diligence = 0
        self._empathy = 0
        self._charm = 0
        self._honour = 0
        self._mute = ""  # Takes lower case true/false but is a varchar (45) and not a Bool

        self.init_stats()  # Assign instance variables

        # Create Inventory object instance via class constructor, using details from Character object instance
        self._inventory = self.init_inventory()

        # Create Account object instance via class constructor, using details from Character object instance
        self._account = self.init_account()

    # fill with attributes from init
    def init_stats(self):
        """Given a dictionary of stats from AzureMS's DB we add them to Character object's attributes

        Runs near the end of Character::__init__(char_stats, database_config).
        It assigns the character attributes in char_stats to their respective protected attributes belonging to
        the Character object instance.
        """
        self._character_id = self._stats['id']
        self._account_id = self._stats['accountid']
        self._name = self._stats['name']
        self._level = self._stats['level']
        self._exp = self._stats['exp']
        self._strength = self._stats['str']
        self._dex = self._stats['dex']
        self._luk = self._stats['luk']
        self._inte = self._stats['int']
        self._max_hp = self._stats['maxhp']
        self._max_mp = self._stats['maxmp']
        self._meso = self._stats['meso']
        self._job = self._stats['job']
        self._skin = self._stats['skincolor']
        self._gender = self._stats['gender']
        self._fame = self._stats['fame']
        self._hair = self._stats['hair']
        self._face = self._stats['face']
        self._ap = self._stats['ap']
        self._map = self._stats['map']
        self._bl_slots = self._stats['buddyCapacity']
        self._rebirths = self._stats['reborns']
        self._ambition = self._stats['ambition']
        self._insight = self._stats['insight']
        self._willpower = self._stats['willpower']
        self._diligence = self._stats['diligence']
        self._empathy = self._stats['empathy']
        self._charm = self._stats['charm']
        self._honour = self._stats['innerExp']  # Best guess - might be wrong!
        self._mute = self._stats['chatban']

    def init_account(self):
        """Fetch a dictionary of account attributes from AzureMS's DB and use it to instantiate a new Account object

        Runs at the end of Character::__init__(char_stats, database_config).
        Checks the account ID associated with the character instance, and uses the Account class constructor to create
        a new Account object instance, with the relevant attributes from the database.

        Returns:
            Account object with attributes identical to its corresponding entry in the database
        Raises:
            Generic error on failure - handled by the utility.get_db_first_hit() method
        """
        account_id = utils.get_db_first_hit(
            self._database_config,
            f"SELECT * FROM characters WHERE id = '{self.character_id}'"
        ).get("accountid")  # get_db() returns a Dictionary, so get() is used to fetch only the value
        # The row will always be 0 because there should be no characters with the same character ID (Primary Key)

        account_info = utils.get_db_first_hit(
            self._database_config,
            f"SELECT * FROM accounts WHERE id = '{account_id}'"
        )  # The row will always be 0 because there should be no characters with the same account ID (Primary Key)

        account = Account(account_info, self.database_config)
        return account

    def init_inventory(self):
        """Fetch a dictionary of user attributes from AzureMS's DB and use it to instantiate a new (custom) Inventory object

        Runs near the end of Character::__init__(char_stats, database_config).
        Uses the Character ID associated with the character instance, and the Inventory class constructor to create
        a new Inventory object instance, with the relevant inventory attributes from the database.

        Raises:
            Generic error on failure - handled by the utility.get_db_first_hit() method
        """
        inventory = Inventory(self.character_id, self.database_config)
        return inventory

    @property
    def database_config(self):
        return self._database_config

    @property
    def stats(self):
        return self._stats

    @property
    def character_id(self):
        return self._character_id  # Only getter, no setter; Primary Key must not be set manually!

    @property
    def account_id(self):
        return self._account_id  # Only getter, no setter; Primary Key must not be set manually!

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, x):
        if x > 275:
            raise ValueError("Level should not exceed 275!")
        else:
            self.set_stat_by_column("level", x)
            self._level = x

    def add_level(self, amount):
        """Adds the specified amount to the current level count

        Args:
            amount: Int, representing the number of levels to be added to the current count
        """
        new_level = int(self.level) + amount
        if new_level > 275:
            raise ValueError("Level should not exceed 275!")
        else:
            self.level = new_level

    @property
    def job(self):
        return self._job

    @job.setter
    def job(self, job_id):
        if str(job_id) not in JOBS:
            raise ValueError("Invalid Job ID!")
        else:
            self.set_stat_by_column("job", job_id)
            self._job = job_id

    def get_job_name(self):
        """Returns the actual name of the job from job id

        Returns:
            String, representing the job name corresponding to a job ID
        """
        return JOBS[str(self.job)]

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name):
        """Set a new name for the character

        Args:
            new_name: string, representing the new character name that will be set in the database
        """
        if len(new_name) > 13:
            raise ValueError("Character names can only be 13 characters long!")
        else:
            self.set_stat_by_column("name", new_name)
            self._name = new_name

    @property
    def meso(self):
        return self._meso

    @meso.setter
    def meso(self, amount):
        if amount > 10000000000:
            raise ValueError("You should not try to set meso to more than 10b!")
        else:
            self.set_stat_by_column("meso", amount)
            self._meso = amount

    def add_mesos(self, amount):
        """Adds the specified amount to the current meso count

        Args:
            amount: Int, representing the amount of mesos to be added to the current count
        """
        new_amount = int(self.meso) + amount
        if new_amount > 10000000000:
            raise ValueError("You should not try to set meso to more than 10b!")
        else:
            self.meso = str(new_amount)  # money is a String; converting back to String for consistency

    @property
    def fame(self):
        return self._fame

    @fame.setter
    def fame(self, amount):
        if amount > 32767:
            raise ValueError("You should not try to set fame to more than 30k!")
        else:
            self.set_stat_by_column("fame", amount)
            self._fame = amount

    def add_fame(self, amount):
        """Adds the specified amount to the current fame count

        Args:
            amount: Int, representing the number of fames to be added to the current count
        """
        new_fame = int(self.fame) + amount
        if new_fame > 32767:
            raise ValueError("You should not try to set fame to more than 30k!")
        else:
            self.fame = new_fame

    @property
    def map(self):
        return self._map

    @map.setter
    def map(self, map_id):
        if map_id < 100000000 or map_id > 999999999:  # Best guess - might be wrong!
            raise ValueError("Wrong map ID!")
        else:
            self.set_stat_by_column("map", map_id)
            self._map = map_id

    @property
    def face(self):
        return self._face

    @face.setter
    def face(self, face_id):
        self.set_stat_by_column("face", face_id)  # TODO: Add check
        self._face = face_id

    @property
    def hair(self):
        return self._hair

    @hair.setter
    def hair(self, hair_id):
        self.set_stat_by_column("hair", hair_id)  # TODO: Add check
        self._hair = hair_id

    @property
    def skin(self):
        return self._skin

    @skin.setter
    def skin(self, skin_id):
        self.set_stat_by_column("skin", skin_id)  # TODO: Add check
        self._skin = skin_id

    @property
    def gender(self):
        return self._gender

    @gender.setter
    def gender(self, gender_id):
        self.set_stat_by_column("gender", gender_id)  # TODO: Add check
        self._gender = gender_id

    @property
    def exp(self):
        return self._exp

    @exp.setter
    def exp(self, exp_amount):
        if exp_amount > 2147483647:
            raise ValueError("You should not try to set EXP above 2bil!")
        else:
            self.set_stat_by_column("exp", exp_amount)
            self._exp = exp_amount

    def add_exp(self, amount):
        """Add the specified amount to the current existing EXP pool

        Args:
            amount: Int, representing the amount of EXP to be added to the current pool
        """
        if amount > 2147483647:
            raise ValueError("You should not try to increment EXP by more than 2bil!")
        else:
            new_exp = int(self.exp) + amount
            self.exp = str(new_exp)  # EXP is a String; converting back to String for consistency

    @property
    def strength(self):
        return self._strength

    @strength.setter
    def strength(self, amount):
        if amount > 32767:
            raise ValueError("You should not try to set STR above 30k!")
        else:
            self.set_stat_by_column("str", amount)
            self._strength = amount

    def add_str(self, amount):
        """Add the specified amount to the current existing STR pool

        Args:
            amount: Int, representing the amount of STR to be added to the current pool
        """
        new_str = int(self.strength) + amount
        if new_str > 32767:
            raise ValueError("You should not try to set STR above 30k!")
        else:
            self.strength = new_str

    @property
    def dex(self):
        return self._dex

    @dex.setter
    def dex(self, amount):
        if amount > 32767:
            raise ValueError("You should not try to set DEX above 30k!")
        else:
            self.set_stat_by_column("dex", amount)
            self._dex = amount

    def add_dex(self, amount):
        """Add the specified amount to the current existing DEX pool

        Args:
            amount: Int, representing the amount of DEX to be added to the current pool
        """
        new_dex = int(self.dex) + amount
        if new_dex > 32767:
            raise ValueError("You should not try to set DEX above 30k!")
        else:
            self.dex = new_dex

    @property
    def inte(self):
        return self._inte

    @inte.setter
    def inte(self, amount):
        if amount > 32767:
            raise ValueError("You should not try to set INT above 30k!")
        else:
            self.set_stat_by_column("int", amount)
            self._inte = amount

    def add_inte(self, amount):
        """Add the specified amount to the current existing INT pool

        Args:
            amount: Int, representing the amount of INT to be added to the current pool
        """
        new_inte = int(self.inte) + amount
        if new_inte > 32767:
            raise ValueError("You should not try to set INT above 30k!")
        else:
            self.inte = new_inte

    @property
    def luk(self):
        return self._luk

    @luk.setter
    def luk(self, amount):
        if amount > 32767:
            raise ValueError("You should not try to set LUK above 30k!")
        else:
            self.set_stat_by_column("luk", amount)
            self._luk = amount

    def add_luk(self, amount):
        """Add the specified amount to the current existing LUK pool

        Args:
            amount: Int, representing the amount of LUK to be added to the current pool
        """
        new_luk = int(self.luk) + amount
        if new_luk > 32767:
            raise ValueError("You should not try to set LUK above 30k!")
        else:
            self.luk = new_luk

    def get_primary_stats(self):
        """Returns str, int, dex, luk values in a dictionary

        Returns:
            dictionary of primary stats
        """
        primary_stats = {
            "str": self.strength,
            "dex": self.dex,
            "int": self.inte,
            "luk": self.luk
        }
        return primary_stats

    @property
    def max_hp(self):
        return self._max_hp

    @max_hp.setter
    def max_hp(self, amount):
        self.set_stat_by_column("maxhp", amount)
        self._max_hp = amount

    def add_max_hp(self, amount):
        """Add the specified amount to the current existing Max HP pool

        Args:
            amount: Int, representing the amount of Max HP to be added to the current pool
        """
        new_hp = int(self.max_hp) + amount
        self.max_hp = new_hp

    @property
    def max_mp(self):
        return self._max_mp

    @max_mp.setter
    def max_mp(self, amount):
        self.set_stat_by_column("maxmp", amount)
        self._max_mp = amount

    def add_max_mp(self, amount):
        """Add the specified amount to the current existing Max MP pool

        Args:
            amount: Int, representing the amount of max MP to be added to the current pool
        """
        new_mp = int(self.max_mp) + amount
        self.max_mp = new_mp

    @property
    def ap(self):
        return self._ap

    @ap.setter
    def ap(self, amount):
        if amount > 32767:
            raise ValueError("You should not try to set AP above 30k!")
        else:
            self.set_stat_by_column("ap", amount)
            self._ap = amount

    def add_ap(self, amount):
        """Add the specified amount to the current existing free AP pool

        Args:
            amount: Int, representing the amount of free AP to be added to the current pool
        """
        new_ap = int(self.ap) + amount
        if new_ap > 32767:
            raise ValueError("You should not try to set AP above 30k!")
        else:
            self.ap = new_ap

    @property
    def bl_slots(self):
        return self._bl_slots

    @bl_slots.setter
    def bl_slots(self, amount):
        # TODO: Add checks
        self.set_stat_by_column("buddyCapacity", amount)
        self._bl_slots = amount

    def add_bl_slots(self, amount):
        # TODO: Add checks
        new_amount = int(self.bl_slots) + amount
        self.bl_slots = new_amount

    @property
    def rebirths(self):
        return self._rebirths

    @rebirths.setter
    def rebirths(self, amount):
        # TODO: Add checks
        self.set_stat_by_column("reborns", amount)
        self._rebirths = amount

    def add_rebirths(self, amount):
        # TODO: Add checks
        new_amount = int(self.rebirths) + amount
        self.rebirths = new_amount

    @property
    def ambition(self):
        return self._ambition

    @ambition.setter
    def ambition(self, amount):
        # TODO: Add checks
        self.set_stat_by_column("ambition", amount)
        self._ambition = amount

    def add_ambition(self, amount):
        # TODO: Add checks
        new_amount = int(self.ambition) + amount
        self.ambition = new_amount

    @property
    def insight(self):
        return self._insight

    @insight.setter
    def insight(self, amount):
        # TODO: Add checks
        self.set_stat_by_column("insight", amount)
        self._insight = amount

    def add_insight(self, amount):
        # TODO: Add checks
        new_amount = int(self.insight) + amount
        self.insight = new_amount

    @property
    def willpower(self):
        return self._willpower

    @willpower.setter
    def willpower(self, amount):
        # TODO: Add checks
        self.set_stat_by_column("willpower", amount)
        self._willpower = amount

    def add_willpower(self, amount):
        # TODO: Add checks
        new_amount = int(self.willpower) + amount
        self.willpower = new_amount

    @property
    def diligence(self):
        return self._diligence

    @diligence.setter
    def diligence(self, amount):
        # TODO: Add checks
        self.set_stat_by_column("diligence", amount)
        self._diligence = amount

    def add_diligence(self, amount):
        # TODO: Add checks
        new_amount = int(self.diligence) + amount
        self.diligence = new_amount

    @property
    def empathy(self):
        return self._empathy

    @empathy.setter
    def empathy(self, amount):
        # TODO: Add checks
        self.set_stat_by_column("empathy", amount)
        self._empathy = amount

    def add_empathy(self, amount):
        # TODO: Add checks
        new_amount = int(self.empathy) + amount
        self.empathy = new_amount

    @property
    def charm(self):
        return self._charm

    @charm.setter
    def charm(self, amount):
        # TODO: Add checks
        self.set_stat_by_column("charm", amount)
        self._charm = amount

    def add_charm(self, amount):
        # TODO: Add checks
        new_amount = int(self.charm) + amount
        self.charm = new_amount

    @property
    def honour(self):
        return self._honour

    @honour.setter
    def honour(self, amount):
        # TODO: Add checks
        self.set_stat_by_column("honour", amount)
        self._honour = amount

    def add_honour(self, amount):
        # TODO: Add checks
        new_amount = int(self.honour) + amount
        self.honour = new_amount

    @property
    def mute(self):
        return self._mute

    @mute.setter
    def mute(self, status):
        # TODO: Add checks
        self.set_stat_by_column("mute", status)
        self._mute = status

    @property
    def account(self):
        return self._account

    @property
    def inventory(self):
        return self._inventory

    def get_char_img(self):
        """Generates character avatar using MapleStory.io API

        Returns:
            url: String, a link to the generated avatar
        """
        equipped_items = [self.face, self.hair]
        equipped_inv = self.inventory.equipped_inv

        for item in equipped_inv:
            item_id = equipped_inv[item]["itemid"]
            equipped_items.append(item_id)

        url = f"https://maplestory.io/api/GMS/216/Character/200{self.skin}/{str(equipped_items)[1:-1]}/stand1/1".replace(" ", "")

        return url

    def set_stat_by_column(self, column, value):
        """Update a character's stats from column name in database

        Grabs the database attributes provided through the class constructor.
        Uses these attributes to attempt a database connection.
        Attempts to update the field represented by the provided column in characterstats, with the provided value.
        Not recommended to use this alone, as it won't update the character object which this was used from.

        Args:
            value: int or string, representing the value to be set in the database
            column: string, representing the column in the database that is to be updated

        Returns:
            A boolean representing whether the operation was successful

        Raises:
            SQL Error 2003: Can't cannect to DB
            WinError 10060: No response from DB
            List index out of range: Wrong column name
        """

        host = self._database_config["host"]
        user = self._database_config["user"]
        password = self._database_config["password"]
        schema = self._database_config["schema"]
        port = self._database_config["port"]

        try:
            database = con.connect(host=host, user=user, password=password, database=schema, port=port)

            cursor = database.cursor(dictionary=True)
            cursor.execute(f"UPDATE characters SET {column} = '{value}' WHERE name = '{self.name}'")
            database.commit()
            print(f"Successfully updated {column} value for character: {self.name}.")
            self._stats[column] = value  # Update the stats in the dictionary
            database.disconnect()
            return True
        except Exception as e:
            print("[ERROR] Error trying to set stats in database.", e)
            return False

    def get_stat_by_column(self, column):
        """Given a column name, return its value in the database

        Args:
            column: string, representing the column in the database from which the value is to be fetched from

        Returns:
            string, representing the value in the database associated with the provided column

        Raises:
            Generic error on failure
        """
        try:
            return self.stats[column]
        except Exception as e:
            print("[ERROR] Error trying to get stats from given column.", e)
            return False



"""This module holds the utility functions and constants for the lazuli package.

Copyright 2020 TEAM SPIRIT. All rights reserved.
Use of this source code is governed by a AGPL-style license that can be found in the LICENSE file.
Refer to database.py or the project wiki on GitHub for usage examples.
"""
import mysql.connector as con

# Dictionary that maps inventory tabs' names to their corresponding index in the DB/source
map_inv_types = {
    'equipped': -1,
    'equip': 1,
    'eqp': 1,
    'use': 2,  # Default name for Lazuli purposes
    'consume': 2,  # name in WZ
    'etc': 4,
    'setup': 3,  # Default name for Lazuli purposes
    'install': 3,  # name in WZ
    'cash': 5
}


def get_key(dictionary, val):
    """Generic function to return the key for a given value

    Iterates through the dictionary, comparing values to see if it matches the desired value.
    If so, return the corresponding key. If no matches are found by the end, return False.
    This function short-circuits (i.e. returns with the first match found).
    Note: OrderedDict is no longer necessary for this as of Python 3.6, as order is preserved automagically.

    Args:
        dictionary: Dictionary, representing the dictionary to be searched
        val: Var, representing the desired/target value to search for

    Returns:
        Var, representing the corresponding key (if any); defaults to False, if none are found

    Raises:
        Generic error for any failures
    """
    try:
        for key, value in dictionary.items():
            if val == value:
                return key
        print("No corresponding key found")
        return False
    except Exception as e:
        print(f"Unexpected error encountered whilst attempting to perform dictionary search: {e}")
        return False


def get_inv_type_by_name(inv_string):
    inv_type = map_inv_types.get(inv_string)
    return inv_type


def get_inv_name_by_type(inv_type):  # Never used
    inv_name = get_key(map_inv_types, inv_type)
    return inv_name


def get_db_first_hit(config, query):
    """Generic top level function for fetching data (first hit) from DB using the provided DB config and query

    This method assumes that only one result is found - it always defaults to the first result.
    An effort has been made to convert this to a decorator so that it may also be applied to
    Character::set_stat_by_column() & Character::get_user_id(), which ultimately ended in failure.

    Args:
        config: dictionary, representing database config attributes
        query: String, representing SQL query

    Returns:
        String representing the result of the provided SQL query, using the provided DB connection attributes

    Raises:
        SQL Error 2003: Can't cannect to DB
        WinError 10060: No response from DB
        List index out of range: Wrong column name
        Generic error as a final catch-all
    """
    try:
        database = con.connect(
            host=config['host'],
            user=config['user'],
            password=config['password'],
            database=config['schema'],
            port=config['port']
        )
        cursor = database.cursor(dictionary=True)
        cursor.execute(query)
        data = cursor.fetchall()[0]
        database.disconnect()

        return data

    except Exception as e:
        print(f"CRITICAL: Error encountered whilst attempting to connect to the database! \n{e}")


def get_db_all_hits(config, query):
    """Generic top level function for fetching all matching data from DB using the provided DB config and query

    Args:
        config: dictionary, representing database config attributes
        query: String, representing SQL query

    Returns:
        List of objects, representing the result of the provided SQL query, using the provided DB connection attributes

    Raises:
        SQL Error 2003: Can't cannect to DB
        WinError 10060: No response from DB
        List index out of range: Wrong column name
        Generic error as a final catch-all
    """
    try:
        database = con.connect(
            host=config['host'],
            user=config['user'],
            password=config['password'],
            database=config['schema'],
            port=config['port']
        )
        cursor = database.cursor(dictionary=True)
        cursor.execute(query)
        data = cursor.fetchall()
        database.disconnect()

        return data

    except Exception as e:
        print(f"CRITICAL: Error encountered whilst attempting to connect to the database! \n{e}")


def has_item_in_inv_type(inv_type, item_id):
    """Checks whether the particular tab of the inventory has an item

    Generic top level function used by Inventory::has_item_in_XXX() methods, and the
    Inventory::is_equipping() method.
    Iterates through the dictionary of items associated with the specified tab, and check if
    the provided item ID can be found as a value.

    Args:
        inv_type: inventory object, representing inventory tab to search
        item_id: int, representing the ID of the item to search for

    Returns:
        Boolean, representing whether the specified item was found
    """
    for bag_index in inv_type:
        if inv_type[bag_index]['itemid'] == item_id:
            return True
    return False


def extract_name(player_list):
    """Extracts a list of players from SQL data, via the name column

    Args:
        player_list: List of dicts, representing list of all players

    Returns:
        List of Strings, representing player names
    """
    try:
        if not player_list[0]['name']:  # if empty or null; sanity check
            raise RuntimeError

        players = []
        for player in player_list:
            players.append(player['name'])
        return players
    except Exception as e:
        print(f"ERROR: Error encountered whilst attempting extract list of player names!\n{e}")


def extract_name_and_value(player_list, column):
    """Extracts a list of players (and a specified value) from SQL data, via the name column

    Args:
        player_list: List of dicts, representing list of all players
        column: String, representing column name to extract

    Returns:
        List of Tuples, representing player names and their corresponding values (e.g. level)
    """
    try:
        if not player_list[0]['name']:  # if empty or null; sanity check
            raise RuntimeError

        players = []
        for player in player_list:
            players.append((player['name'], player[column]))  # List of Tuples, as per Brandon's request
        return players
    except Exception as e:
        print(f"ERROR: Error encountered whilst attempting extract list of player names!\n{e}")



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Operating System :: OS Independent',
    'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
    'Natural Language :: English',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Topic :: Games/Entertainment :: Side-Scrolling/Arcade Games',
]

setup(
    name="lazuli",
    version="0.0.7",
    description="A Python-based tool for interacting with AzureMSv316-based databases.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TEAM-SPIRIT-Productions/Lazuli",
    author="Amos Chua",
    classifiers=classifiers,
    keywords="database, azure, maplestory",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=["mysql-connector-python-rf"],
)



"""This module holds the Inventory class for the lazuli package.

Copyright 2020 TEAM SPIRIT. All rights reserved.
Use of this source code is governed by a AGPL-style license that can be found in the LICENSE file.
Refer to database.py or the project wiki on GitHub for usage examples.
"""
import mysql.connector as con
import lazuli.utility as utils


class Inventory:
	"""Inventory object; quasi-models AzureMS inventories.

	The instance method Lazuli::get_char_by_name(name) creates a Character object; as part of
	lazuli Character object instantiation, an Inventory object instance containing inventory attributes of
	the character with IGN "name" in the connected AzureMS-based database is created.
	This class contains the appropriate getter methods for said attributes.
	As a consequence of the inherent complexity of MapleStory's item system, for safety reasons,
	this module offers NO inventory-write operations (aka setters).

	Attributes:
		equip_inv: list of dictionaries, representing in-game items contained by the EQUIP tab
		consume_inv: list of dictionaries, representing in-game items contained by the USE tab
		etc_inv: list of dictionaries, representing in-game items contained by the ETC tab
		install_inv: list of dictionaries, representing in-game items contained by the SETUP tab
		cash_inv: list of dictionaries, representing in-game items contained by the CASH tab
		equipped_inv: list of dictionaries, representing in-game items currently equipped by the character
	"""

	def __init__(self, character_id, db_config):
		"""Inventory object; quasi-models AzureMS inventories.

		Modelled after SwordieDB project's Inventory class init method.
		Due to the inherent complexity of MapleStory's inventory system, this Inventory object will
		attempt to contain attributes of all 6 of AzureMS's inventory types, using a custom object.
		Every inventory attribute is a dictionary of dictionaries, the latter of which models the contents of
		the `inventoryitems` table in a AzureMS-based database.
	"""
		self._character_id = character_id
		self._database_config = db_config

		self._equip_inv = self.init_equip_items()
		self._use_inv = self.init_use_inv()
		self._etc_inv = self.init_etc_inv()
		self._cash_inv = self.init_cash_inv()
		self._install_inv = self.init_install_inv()

		self._equipped_inv = self.init_equipped_inv()

	@property
	def database_config(self):
		return self._database_config

	@property
	def character_id(self):
		return self._character_id

	@property
	def equip_inv(self):
		return self._equip_inv

	@property
	def consume_inv(self):
		return self._use_inv

	@property
	def etc_inv(self):
		return self._etc_inv

	@property
	def cash_inv(self):
		return self._cash_inv

	@property
	def install_inv(self):
		return self._install_inv

	@property
	def equipped_inv(self):
		return self._equipped_inv

	def load_inv(self, inv_type):
		"""Given an inventory_type, fetch every item associated with it (i.e. -1, 1, 2, 3, 4, 5)

		Args:
			inv_type: int

		Returns:
			Dictionary of Dictionaries, representing all the in-game items that in the specified inventory type

		Raises:
			Generic error on failure
		"""
		try:
			database = con.connect(
				host=self.database_config['host'],
				user=self.database_config['user'],
				password=self.database_config['password'],
				database=self.database_config['schema'],
				port=self.database_config['port']
			)
			cursor = database.cursor(dictionary=True)
			cursor.execute(
				f"SELECT * FROM inventoryitems WHERE characterid = '{self.character_id}' AND inventorytype = '{inv_type}'")
			inventory = cursor.fetchall()

			inv = {}

			for items in inventory:
				# More to add if needed.
				bag_index = items["position"]
				item_id = items["itemid"]
				quantity = items["quantity"]
				is_cash = items["isCash"]
				inventory_type = items["inventorytype"]
				item_stats = {
					"itemid": item_id,
					"quantity": quantity,  # Never used
					"inventorytype": inventory_type,
					"iscash": is_cash  # Never used
				}
				inv[bag_index] = item_stats
				# Key value, we set as the bag_index aka position of the item in the inventory.
			return inv
		except Exception as e:
			print(f"[ERROR] Error trying to load inventory type {inv_type}", e)

	def init_equip_items(self):
		return self.load_inv(utils.get_inv_type_by_name("equip"))

	def init_use_inv(self):
		return self.load_inv(utils.get_inv_type_by_name("use"))

	def init_etc_inv(self):
		return self.load_inv(utils.get_inv_type_by_name("etc'"))

	def init_cash_inv(self):
		return self.load_inv(utils.get_inv_type_by_name("cash"))

	def init_equipped_inv(self):
		return self.load_inv(utils.get_inv_type_by_name("equipped"))

	def init_install_inv(self):
		return self.load_inv(utils.get_inv_type_by_name("setup"))

	def has_item_in_equip(self, item_id):
		"""Checks whether the EQUIP tab of the inventory has an item
		Uses Inventory::has_item_in_inv_type()
		Returns:
			Boolean, representing whether the specified item was found
		"""
		return utils.has_item_in_inv_type(self.equip_inv, item_id)

	def has_item_in_consume(self, item_id):
		"""Checks whether the USE tab of the inventory has an item
		Uses Inventory::has_item_in_inv_type()
		Returns:
			Boolean, representing whether the specified item was found
		"""
		return utils.has_item_in_inv_type(self.consume_inv, item_id)

	def has_item_in_etc(self, item_id):
		"""Checks whether the ETC tab of the inventory has an item
		Uses Inventory::has_item_in_inv_type()
		Returns:
			Boolean, representing whether the specified item was found
		"""
		return utils.has_item_in_inv_type(self.etc_inv, item_id)

	def has_item_in_install(self, item_id):
		"""Checks whether the SETUP tab of the inventory has an item
		Uses Inventory::has_item_in_inv_type()
		Returns:
			Boolean, representing whether the specified item was found
		"""
		return utils.has_item_in_inv_type(self.install_inv, item_id)

	def has_item_in_cash(self, item_id):
		"""Checks whether the CASH tab of the inventory has an item
		Uses Inventory::has_item_in_inv_type()
		Returns:
			Boolean, representing whether the specified item was found
		"""
		return utils.has_item_in_inv_type(self.cash_inv, item_id)

	def is_equipping(self, item_id):
		"""Checks whether the EQUIP window (i.e. Hotkey "E") has an item (i.e. item is equipped)
		Uses Inventory::has_item_in_inv_type()
		Returns:
			Boolean, representing whether the specified item was found
		"""
		return utils.has_item_in_inv_type(self.equipped_inv, item_id)



"""Lazuli is designed for use in development of AzureMSv316-based MapleStory private server tools (e.g. Discord bots).

Copyright 2020 TEAM SPIRIT. All rights reserved.
Use of this source code is governed by a AGPL-style license that can be found in the LICENSE file.
This module contains the main class that users would instantiate: Lazuli.
Users can use this class to fetch and manipulate information from the database.
Refer to the project wiki on GitHub for more in-depth examples.

    Typical usage example:

    lazuli = Lazuli()  # Instantiate DB object
    char = lazuli.get_char_by_name("KOOKIIE")  # Instantiate Character object
    meso = char.money  # Use of Character methods to fetch data from DB
    char.money = 123456789  # Use of Character methods to write data to DB
"""
import mysql.connector as con
from lazuli.character import Character
from lazuli.account import Account
import lazuli.utility as utils


class Lazuli:
    """Database object; models the AzureMS DB.

    Use this class to create instances of AzureMS characters, or inventories, complete with their respective
    data from the connected AzureMS-based database.
    Using instance method Lazuli::get_char_by_name("name") will create a Character object (see character.py) instance
    that has attributes identical to the character with IGN "name" in the connected AzureMS-based database.

    Attributes:
        host: Optional; IP address of the database. Defaults to "localhost"
        schema: Optional; Name of the schema of the database. Defaults to "kms_316`"
        user: Optional; Username for access to the database. Defaults to "root"
        password: Optional; Password for access to the database. Defaults to ""
        port: Optional; Port with which to access the database. Defaults to 3306
    """

    def __init__(self, host="localhost", schema="kms_316", user="root", password="", port=3306):
        self._host = host
        self._schema = schema
        self._user = user
        self._password = password
        self._port = port

        self._database_config = {
            'host': self.host,
            'user': self.user,
            'password': self.password,
            'schema': self.schema,
            'port': self.port
        }

    @property
    def host(self):
        return self._host

    @host.setter
    def host(self, x):
        self._host = x

    @property
    def schema(self):
        return self._schema

    @schema.setter
    def schema(self, x):
        self._schema = x

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, x):
        self._user = x

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, x):
        self._password = x

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, new_port):
        self._port = new_port

    def get_char_by_name(self, char_name):
        """Create an instance of a Character object from the given character name

        Uses the class constructor of the Character class to create a new instance, with the corresponding
        character data and database attributes from the connected database.

        Args:
            char_name: string, representing character name (aka IGN)

        Returns:
            Character object instantiated with corresponding data from the connected database.
            Defaults to None if the operation fails.

        Raises:
            Generic error on failure - handled by the Character::get_db() method
        """
        character_stats = Character.get_db_first_hit(
            self._database_config,
            f"SELECT * FROM characters WHERE `name` = '{char_name}'"
        )  # Fetch first result because there should only be one character with that name

        character = Character(character_stats, self._database_config)
        return character

    def get_account_by_username(self, username):
        """Given a username (NOT IGN), create a new account object instance

        Fetches the user account attributes from the database by querying for username.
        uses the User class constructor to create a new User object instance, with the said attributes.
        Useful for getting account information from accounts with no characters.

        Args:
            username: String, representing the username used for logging the user into game

        Returns:
            Account object with attributes identical to its corresponding entry in the database

        Raises:
            Generic error on failure - handled by the Character::get_db() method
        """
        account_info = Character.get_db_first_hit(
            self._database_config,
            f"SELECT * FROM accounts WHERE `name` = '{username}'"
        )  # Fetch first result because there should only be one character with that name

        account = Account(account_info, self._database_config)
        return account

    def set_char_stat(self, name, column, value):
        """Given a character name and column name, change its value in the database

        Args:
            column: string, representing the column in the database
            name: string, representing the character name in the database
            value: string/int, representing the value that is to be updated in the corresponding field

        Returns:
            boolean, representing whether the operation completed successfully

        Raises:
            SQL Error 2003: Can't cannect to DB
            WinError 10060: No response from DB
            List index out of range: Wrong column or character name
        """
        try:
            database = con.connect(host=self.host, user=self.user, password=self.password, database=self.schema,
                                   port=self.port)
            cursor = database.cursor(dictionary=True)
            cursor.execute(f"UPDATE characters SET {column} = '{value}' WHERE `name` = '{name}'")
            database.commit()
            database.disconnect()
            print(f"Successfully set {name}'s stats in database.")
            return True
        except Exception as e:
            print("[ERROR] Error trying to update character stats in Database.", e)
            return False

    def get_db_all_hits(self, query):
        """Fetch all matching data from DB using the provided query

        Wrapper function. Uses the DB config from Lazuli attributes for DB connection.
        Feeds the config values into utility.get_db_all_hits().

        Args:
            query: String, representing SQL query

        Returns:
            List of objects, representing the result of the provided SQL query, using the provided DB connection attributes

        Raises:
            Generic error on failure - handled by the utility.get_db_all_hits() method

        """
        data = utils.get_db_all_hits(self._database_config, query)
        return data

    def get_online_list(self):
        """Fetch the list of players currently online

        AzureMS stores login state in the DB, in the 'accounts' table, 'loggedin' column.
        Lazuli::get_online_list queries for a list of all accounts that are logged in,
        using the Lazuli::get_db_all_hits method.

        Returns:
            List, representing all players online. Defaults to False in the event of an error during execution

        Raises:
            Generic error on failure
        """
        data = self.get_db_all_hits(f"SELECT * FROM accounts WHERE `loggedin` > 0")
        return data  # List of online players

    def get_online_count(self):
        """Fetch the number of players currently online

        Uses the Lazuli::get_online_list method to fetch the list of all players online.
        Counts the length of said list, to obtain number of players online.

        Returns:
            Int, representing number of players online. Defaults to False in the event of an error during execution

        Raises:
            Generic error on failure, handled by Lazuli::get_online_list()
        """
        players = self.get_online_list()
        return len(players)

    def get_online_players(self):
        """Fetch usernames of all players currently online

        Uses the Lazuli::get_online_list method to fetch the list of all players online.
        Extract the usernames from the said list.

        Returns:
            List, representing all players online. Defaults to False in the event of an error during execution

        Raises:
            Generic error on failure, handled by Lazuli::get_online_list()
        """
        player_data = self.get_online_list()
        return utils.extract_name(player_data)

    def get_level_ranking(self, number_of_players=5):
        """Fetches the top ranking players in terms of level

        Uses Lazuli::get_db_all_hits to query, and utility.extract_name_and_value to process the data.

        Args:
            number_of_players (optional): Int, e.g. Top 5 Ranking (default), Top 10 Ranking, etc.

        Returns:
            List of Tuples, representing player names and their corresponding level
        """
        player_data = self.get_db_all_hits(f"SELECT * FROM `characters` ORDER BY `level` DESC LIMIT {number_of_players}")
        return utils.extract_name_and_value(player_data, "level")

    def get_meso_ranking(self, number_of_players=5):
        """Fetches the top ranking players in terms of mesos

        Uses Lazuli::get_db_all_hits to query, and utility.extract_name_and_value to process the data.

        Args:
            number_of_players (optional): Int, e.g. Top 5 Ranking (default), Top 10 Ranking, etc.

        Returns:
            List of Tuples, representing player names and their corresponding mesos
        """
        player_data = self.get_db_all_hits(f"SELECT * FROM `characters` ORDER BY `meso` DESC LIMIT {number_of_players}")
        return utils.extract_name_and_value(player_data, "meso")

    def get_fame_ranking(self, number_of_players=5):
        """Fetches the top ranking players in terms of fame

        Uses Lazuli::get_db_all_hits to query, and utility.extract_name_and_value to process the data.

        Args:
            number_of_players (optional): Int, e.g. Top 5 Ranking (default), Top 10 Ranking, etc.

        Returns:
            List of Tuples, representing player names and their corresponding fame
        """
        player_data = self.get_db_all_hits(f"SELECT * FROM `characters` ORDER BY `fame` DESC LIMIT {number_of_players}")
        return utils.extract_name_and_value(player_data, "fame")

    def get_rebirth_ranking(self, number_of_players=5):
        """Fetches the top ranking players in terms of rebirths

        Uses Lazuli::get_db_all_hits to query, and utility.extract_name_and_value to process the data.

        Args:
            number_of_players (optional): Int, e.g. Top 5 Ranking (default), Top 10 Ranking, etc.

        Returns:
            List of Tuples, representing player names and their corresponding rebirths
        """
        player_data = self.get_db_all_hits(f"SELECT * FROM `characters` ORDER BY `reborns` DESC LIMIT {number_of_players}")
        return utils.extract_name_and_value(player_data, "reborns")

    def get_rebirth_ranking_by_job_id(self, job_id, number_of_players=5):
        """Fetches the top ranking players of a particular class (specific Job ID), in terms of rebirths

        Uses Lazuli::get_db_all_hits to query, and utility.extract_name_and_value to process the data.

        Args:
            job_id: Int, representing specific Job ID to query
            number_of_players (optional): Int, e.g. Top 5 Ranking (default), Top 10 Ranking, etc.

        Returns:
            List of Tuples, representing player names and their corresponding rebirths
        """
        player_data = self.get_db_all_hits(f"SELECT * FROM `characters` WHERE `job`={job_id} ORDER BY `reborns` DESC LIMIT {number_of_players}")
        return utils.extract_name_and_value(player_data, "reborns")
